@@ -1,4 +1,4 @@
-/* $Id: VBoxNetFlt-linux.c 73007 2011-07-19 09:58:08Z aleksey $ */
+/* $Id: VBoxNetFlt-linux.c 78331 2012-06-04 09:09:22Z aleksey $ */
 /** @file
  * VBoxNetFlt - Network Filter Driver (Host), Linux Specific Code.
  */
@@ -2064,6 +2064,8 @@ static int vboxNetFltLinuxAttachToInterface(PVBOXNETFLTINS pThis, struct net_dev
     pThis->u.s.PacketType.dev  = pDev;
     pThis->u.s.PacketType.func = vboxNetFltLinuxPacketHandler;
     dev_add_pack(&pThis->u.s.PacketType);
+    ASMAtomicUoWriteBool(&pThis->u.s.fPacketHandler, true);
+    Log(("vboxNetFltLinuxAttachToInterface: this=%p: Packet handler installed.\n", pThis));
 
 #ifdef VBOXNETFLT_WITH_FILTER_HOST2GUEST_SKBS_EXPERIMENT
     vboxNetFltLinuxHookDev(pThis, pDev);
@@ -2091,7 +2093,6 @@ static int vboxNetFltLinuxAttachToInterface(PVBOXNETFLTINS pThis, struct net_dev
         pDev = NULL; /* don't dereference it */
     }
     RTSpinlockReleaseNoInts(pThis->hSpinlock, &Tmp);
-    Log(("vboxNetFltLinuxAttachToInterface: this=%p: Packet handler installed.\n", pThis));
 
     /*
      * If the above succeeded report GSO capabilities,  if not undo and
@@ -2139,6 +2140,7 @@ static int vboxNetFltLinuxAttachToInterface(PVBOXNETFLTINS pThis, struct net_dev
 static int vboxNetFltLinuxUnregisterDevice(PVBOXNETFLTINS pThis, struct net_device *pDev)
 {
     RTSPINLOCKTMP Tmp = RTSPINLOCKTMP_INITIALIZER;
+    bool fRegistered;
 
     Assert(!pThis->fDisconnectedFromHost);
 
@@ -2149,26 +2151,37 @@ static int vboxNetFltLinuxUnregisterDevice(PVBOXNETFLTINS pThis, struct net_devi
     vboxNetFltLinuxQdiscRemove(pThis, pDev);
 #endif /* VBOXNETFLT_WITH_QDISC */
 
+    if (ASMAtomicCmpXchgBool(&pThis->u.s.fPacketHandler, false, true))
+    {
+        dev_remove_pack(&pThis->u.s.PacketType);
+        Log(("vboxNetFltLinuxUnregisterDevice: this=%p: packet handler removed.\n", pThis));
+    }
+
     RTSpinlockAcquireNoInts(pThis->hSpinlock, &Tmp);
-    ASMAtomicWriteBool(&pThis->u.s.fRegistered, false);
-    ASMAtomicWriteBool(&pThis->fDisconnectedFromHost, true);
-    ASMAtomicUoWriteNullPtr(&pThis->u.s.pDev);
+    fRegistered = ASMAtomicXchgBool(&pThis->u.s.fRegistered, false);
+    if (fRegistered)
+    {
+        ASMAtomicWriteBool(&pThis->fDisconnectedFromHost, true);
+        ASMAtomicUoWriteNullPtr(&pThis->u.s.pDev);
+    }
     RTSpinlockReleaseNoInts(pThis->hSpinlock, &Tmp);
 
-    dev_remove_pack(&pThis->u.s.PacketType);
+    if (fRegistered)
+    {
 #ifndef VBOXNETFLT_LINUX_NO_XMIT_QUEUE
-    skb_queue_purge(&pThis->u.s.XmitQueue);
+        skb_queue_purge(&pThis->u.s.XmitQueue);
 #endif
-    Log(("vboxNetFltLinuxUnregisterDevice: this=%p: Packet handler removed, xmit queue purged.\n", pThis));
-    Log(("vboxNetFltLinuxUnregisterDevice: Device %p(%s) released. ref=%d\n",
-         pDev, pDev->name,
+        Log(("vboxNetFltLinuxUnregisterDevice: this=%p: xmit queue purged.\n", pThis));
+        Log(("vboxNetFltLinuxUnregisterDevice: Device %p(%s) released. ref=%d\n",
+             pDev, pDev->name,
 #if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 37)
-         netdev_refcnt_read(pDev)
+             netdev_refcnt_read(pDev)
 #else
-         atomic_read(&pDev->refcnt)
+             atomic_read(&pDev->refcnt)
 #endif
-         ));
-    dev_put(pDev);
+           ));
+        dev_put(pDev);
+    }
 
     return NOTIFY_OK;
 }
@@ -2412,7 +2425,11 @@ int vboxNetFltOsDisconnectIt(PVBOXNETFLTINS pThis)
      * Remove packet handler when we get disconnected from internal switch as
      * we don't want the handler to forward packets to disconnected switch.
      */
-    dev_remove_pack(&pThis->u.s.PacketType);
+    if (ASMAtomicCmpXchgBool(&pThis->u.s.fPacketHandler, false, true))
+    {
+        dev_remove_pack(&pThis->u.s.PacketType);
+        Log(("vboxNetFltOsDisconnectIt: this=%p: Packet handler removed.\n", pThis));
+    }
     return VINF_SUCCESS;
 }
 
@@ -2459,7 +2476,7 @@ void vboxNetFltOsDeleteInstance(PVBOXNETFLTINS pThis)
 
     RTSpinlockAcquireNoInts(pThis->hSpinlock, &Tmp);
     pDev = ASMAtomicUoReadPtrT(&pThis->u.s.pDev, struct net_device *);
-    fRegistered = ASMAtomicUoReadBool(&pThis->u.s.fRegistered);
+    fRegistered = ASMAtomicXchgBool(&pThis->u.s.fRegistered, false);
     RTSpinlockReleaseNoInts(pThis->hSpinlock, &Tmp);
 
     if (fRegistered)
@@ -2469,7 +2486,7 @@ void vboxNetFltOsDeleteInstance(PVBOXNETFLTINS pThis)
 #ifndef VBOXNETFLT_LINUX_NO_XMIT_QUEUE
         skb_queue_purge(&pThis->u.s.XmitQueue);
 #endif
-        Log(("vboxNetFltOsDeleteInstance: this=%p: Packet handler removed, xmit queue purged.\n", pThis));
+        Log(("vboxNetFltOsDeleteInstance: this=%p: xmit queue purged.\n", pThis));
         Log(("vboxNetFltOsDeleteInstance: Device %p(%s) released. ref=%d\n",
              pDev, pDev->name,
 #if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 37)
@@ -2516,8 +2533,9 @@ int  vboxNetFltOsPreInitInstance(PVBOXNETFLTINS pThis)
      * Init the linux specific members.
      */
     ASMAtomicUoWriteNullPtr(&pThis->u.s.pDev);
-    pThis->u.s.fRegistered = false;
+    pThis->u.s.fRegistered     = false;
     pThis->u.s.fPromiscuousSet = false;
+    pThis->u.s.fPacketHandler  = false;
     memset(&pThis->u.s.PacketType, 0, sizeof(pThis->u.s.PacketType));
 #ifndef VBOXNETFLT_LINUX_NO_XMIT_QUEUE
     skb_queue_head_init(&pThis->u.s.XmitQueue);
